use crate::params::{k, m, n, o, Level};
use std::ops::{Add, Neg};

#[derive(Debug, Clone, Copy)]
pub struct Mat<const R: usize, const C: usize>(pub [[u8; C]; R]);

// TODO: Bitslicing (not u8's) (see below) (fix this so KATs verify)
// TODO: Finish Simd optimization
// TODO: convert these to Rayon: put loops over things into mp
// TODO: Optimize matrix Mul: Strassen or smth

const MUL_LOOKUP: [[u8; 16]; 16] = [
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
    [0, 2, 4, 6, 8, 10, 12, 14, 3, 1, 7, 5, 11, 9, 15, 13],
    [0, 3, 6, 5, 12, 15, 10, 9, 11, 8, 13, 14, 7, 4, 1, 2],
    [0, 4, 8, 12, 3, 7, 11, 15, 6, 2, 14, 10, 5, 1, 13, 9],
    [0, 5, 10, 15, 7, 2, 13, 8, 14, 11, 4, 1, 9, 12, 3, 6],
    [0, 6, 12, 10, 11, 13, 7, 1, 5, 3, 9, 15, 14, 8, 2, 4],
    [0, 7, 14, 9, 15, 8, 1, 6, 13, 10, 3, 4, 2, 5, 12, 11],
    [0, 8, 3, 11, 6, 14, 5, 13, 12, 4, 15, 7, 10, 2, 9, 1],
    [0, 9, 1, 8, 2, 11, 3, 10, 4, 13, 5, 12, 6, 15, 7, 14],
    [0, 10, 7, 13, 14, 4, 9, 3, 15, 5, 8, 2, 1, 11, 6, 12],
    [0, 11, 5, 14, 10, 1, 15, 4, 7, 12, 2, 9, 13, 6, 8, 3],
    [0, 12, 11, 7, 5, 9, 14, 2, 10, 6, 1, 13, 15, 3, 4, 8],
    [0, 13, 9, 4, 1, 12, 8, 5, 2, 15, 11, 6, 3, 14, 10, 7],
    [0, 14, 15, 1, 13, 3, 2, 12, 9, 7, 6, 8, 4, 10, 11, 5],
    [0, 15, 13, 2, 9, 6, 4, 11, 1, 14, 12, 3, 8, 7, 5, 10],
];

#[inline]
pub fn mul(f16_1: u8, f16_2: u8) -> u8 {
    MUL_LOOKUP[f16_1 as usize][f16_2 as usize]
}

pub fn upper<const N: usize>(mat: Mat<N, N>) -> Mat<N, N> {
    let mut up = Mat([[0; N]; N]);
    for i in 0..N {
        for j in (i + 1)..N {
            up.write(i, j, mat.access(i, j) + mat.access(j, i));
        }
        for i in 0..N {
            up.write(i, i, mat.access(i, i));
        }
    }
    up
}

impl<const R: usize, const C: usize> Mat<R, C> {
    #[inline]
    pub fn access(&self, i: usize, j: usize) -> u8 {
        self.0[i][j]
    }

    #[inline]
    pub fn write(&mut self, i: usize, j: usize, v: u8) {
        self.0[i][j] = v;
    }

    pub fn tp(&self) -> Mat<C, R> {
        let mut tp = Mat([[0; R]; C]);
        for i in 0..R {
            for j in 0..C {
                tp.write(j, i, self.access(i, j));
            }
        }
        tp
    }

    pub fn mul_right<const N: usize>(&self, other: Mat<C, N>) -> Mat<R, N> {
        let other_tp = other.tp();
        let mut out = Mat([[0; N]; R]);
        for i in 0..R {
            for j in 0..N {
                let write = (0..C)
                    .map(|k| mul(other_tp.0[j][k], self.0[i][k]))
                    .fold(0, |acc, x| acc ^ x);
                out.write(i, j, write);
            }
        }
        out
    }
}

impl<const R: usize, const C: usize> Add for Mat<R, C> {
    type Output = Mat<R, C>;

    fn add(self, rhs: Self) -> Self::Output {
        let mut out = Mat([[0; C]; R]);
        for i in 0..R {
            for j in 0..C {
                out.write(i, j, self.access(i, j) ^ rhs.access(i, j))
            }
        }
        out
    }
}

impl<const R: usize, const C: usize> Neg for Mat<R, C> {
    type Output = Mat<R, C>;

    fn neg(self) -> Self::Output {
        self
    }
}

pub fn decode_matrix<const R: usize, const C: usize, const L: Level>(bs: &[u8]) -> Mat<R, C> {
    let mut mat = Mat([[0; C]; R]);
    for i in 0..R {
        for j in 0..C {
            if (j + C * i) % 2 == 0 {
                mat.write(i, j, (bs[(j + C * i) / 2] >> 4) & 0xf);
            } else {
                mat.write(i, j, bs[(j + C * i) / 2] & 0xf);
            }
        }
    }
    mat
}

pub fn decode_matrices<const R: usize, const C: usize, const M: usize, const L: Level>(
    bs: &[u8],
    num: usize,
    triangular: bool,
) -> [Mat<R, C>; M] {
    let mut m = [Mat([[0; C]; R]); M];
    match triangular {
        true => {
            let mut ind = 0;
            for s in 0..M {
                for i in 0..R {
                    for j in i..C {
                        if ind % 2 == 0 {
                            m[s].write(i, j, (bs[ind / 2] >> 4) & 0xf);
                        } else {
                            m[s].write(i, j, bs[ind / 2] & 0xf);
                        }
                        ind += 1;
                    }
                }
            }
        }
        false => {
            for s in 0..M {
                for i in 0..R {
                    for j in 0..C {
                        if (j + C * i) % 2 == 0 {
                            m[s].write(i, j, (bs[(j + C * i + R * C * s) / 2] >> 4) & 0xf);
                        } else {
                            m[s].write(i, j, bs[(j + C * i + R * C * s) / 2] & 0xf);
                        }
                    }
                }
            }
        }
    }
    m
}

pub fn encode_matrices<const R: usize, const C: usize, const M: usize, const L: Level>(
    bs: &[Mat<R, C>],
    triangular: bool,
) -> Vec<u8> {
    let mut outvec = vec![];
    match triangular {
        true => {
            let mut ind = 0;
            for mat in bs {
                for i in 0..R {
                    for j in i..C {
                        if ind % 2 == 0 {
                            outvec.push(mat.access(i, j) << 4);
                        } else {
                            outvec[ind / 2] |= mat.access(i, j);
                        }
                        ind += 1;
                    }
                }
            }
        }
        false => {
            for mat in bs {
                for i in 0..R {
                    for j in 0..C {
                        if (j + C * i) % 2 == 0 {
                            outvec.push(mat.access(i, j) << 4);
                        } else {
                            outvec[(j + C * i) / 2] |= mat.access(i, j);
                        }
                    }
                }
            }
        }
    }
    outvec
}

pub fn decode_vec<const R: usize>(bs: &[u8]) -> Mat<R, 1> {
    let mut ret = Mat([[0; 1]; R]);
    for i in 0..((R + 1) / 2) {
        ret.write(2 * i, 0, (bs[i] >> 4) & 0xf);
        if 2 * i + 1 < R {
            ret.write(2 * i + 1, 0, bs[i] & 0xf);
        }
    }
    ret
}

pub fn encode_vec<const R: usize>(vec: Mat<R, 1>) -> [u8; (R + 1) / 2] {
    let mut out = [0; (R + 1) / 2];
    for i in 0..((R + 1) / 2) {
        out[i] = vec.access(2 * i, 0) << 4;
        if 2 * i + 1 < R {
            out[i] |= vec.access(2 * i + 1, 0);
        }
    }
    out
}

pub fn sample_solution<const L: Level>(
    mat: Mat<{ m(L) }, { k(L) * o(L) }>,
    y: Mat<{ m(L) }, 1>,
    r: Mat<{ k(L) * o(L) }, 1>,
) -> Mat<{ n(L) }, 1> {
    todo!()
}

// pub struct FullMat<const R: usize, const C: usize>([[u8; C]; R]);
//
// pub const fn tpn(n: usize) -> usize {
//     (n * (n + 1)) / 2
// }
//
// pub struct TMat<const N: usize>([u8; tpn(N)])
// where
//     [(); tpn(N)]:;
//
// pub enum Mat<const R: usize, const C: usize>
// where
//     [(); tpn(R)]:,
// {
//     Full(FullMat<R, C>),
//     Upper(TMat<R>),
// }
//
// pub fn upper<const N: usize>(mat: Mat<N, N>) -> Mat<N, N>
// where
//     [(); tpn(N)]:,
// {
//     match mat {
//         Mat::Full(f) => {
//             let mut up = Mat::Upper(TMat([0; tpn(N)]));
//             for i in 0..N {
//                 for j in (i + 1)..N {
//                     up.write(i, j, mat.access(i, j) + mat.access(j, i));
//                 }
//             }
//             for i in 0..N {
//                 up.write(i, i, mat.access(i, i));
//             }
//             up
//         }
//         Mat::Upper(u) => mat,
//     }
// }
//
// impl<const R: usize, const C: usize> Mat<R, C>
// where
//     [(); tpn(R)]:,
//     [(); tpn(C)]:,
// {
//     pub fn access(&self, i: usize, j: usize) -> u8 {
//         match self {
//             Mat::Full(f) => f.0[i][j],
//             Mat::Upper(u) => {
//                 if i > j {
//                     0
//                 } else {
//                     u.0[tpn(R) - tpn(R - i) + j - i]
//                 }
//             }
//         }
//     }
//
//     pub fn write(&mut self, i: usize, j: usize, v: u8) {
//         match self {
//             Mat::Full(f) => {
//                 f.0[i][j] = v;
//             }
//             Mat::Upper(u) => {
//                 if i <= j {
//                     u.0[tpn(R) - tpn(R - i) + j - i] = v;
//                 }
//             }
//         }
//     }
//
//     pub fn tp(&self) -> Mat<C, R> {
//         todo!()
//     }
// }
//
// pub fn decode_matrix<const R: usize, const C: usize, const L: Level>(bs: &[u8]) -> Mat<R, C>
// where
//     [(); tpn(R)]:,
// {
//     todo!()
// })
//
// pub fn decode_matrices<const R: usize, const C: usize, const M: usize, const L: Level>(
//     bs: &[u8],
//     num: usize,
//     triangular: bool,
// ) -> [Mat<R, C>; M]
// where
//     [(); tpn(R)]:,
// {
//     todo!()
// }

#[cfg(test)]
mod test {
    use crate::math::{upper, Mat};
    #[test]
    fn test_upper() {
        let m = Mat([[0, 1, 2], [3, 4, 5], [6, 7, 8]]);
        println!("{:?}", upper(m));
        assert!(true);
    }
}
