use crate::math::{
    decode_matrices, decode_matrix, decode_vec, encode_matrices, encode_vec, sample_solution,
    upper, Mat,
};
use crate::params::Level;
use crate::params::{
    digest_bytes, k, l_bytes, m, n, o, o_bytes, p1_bytes, p2_bytes, p3_bytes, pk_seed_bytes,
    polymat, salt_bytes, v_bytes,
};
use crate::sig::{KeyGen, Signer, Verifier};
use crate::sym::{aes_128_ctr, shake_256};
use rand::rngs::OsRng;
use rand::RngCore;
use rayon::prelude::*;

#[derive(Debug)]
pub struct SigningKey<const L: Level>([u8; salt_bytes(L)], VerifyKey<L>)
where
    [(); salt_bytes(L)]:,
    [(); pk_seed_bytes(L) + p3_bytes(L)]:;

#[derive(Debug)]
pub struct VerifyKey<const L: Level>([u8; pk_seed_bytes(L) + p3_bytes(L)])
where
    [(); pk_seed_bytes(L) + p3_bytes(L)]:;

impl<const L: Level> KeyGen for SigningKey<L>
where
    [(); salt_bytes(L)]:,
    [(); pk_seed_bytes(L) + p3_bytes(L)]:,
    [(); o_bytes(L) + pk_seed_bytes(L)]:,
    [(); n(L) - o(L)]:,
    [(); m(L)]:,
{
    fn gen() -> Self {
        let mut seed_sk: [u8; salt_bytes(L)] = [0; salt_bytes(L)];
        OsRng.fill_bytes(&mut seed_sk);

        let s: [u8; o_bytes(L) + pk_seed_bytes(L)] = shake_256(vec![&seed_sk]);

        let seed_pk = &s[0..pk_seed_bytes(L)];

        let o_bs = &s[pk_seed_bytes(L)..(pk_seed_bytes(L) + o_bytes(L))];
        let om: Mat<{ n(L) - o(L) }, { o(L) }> =
            decode_matrix::<{ n(L) - o(L) }, { o(L) }, { L }>(o_bs);

        let p = aes_128_ctr(seed_pk, p1_bytes(L) + p2_bytes(L));

        let p1: [Mat<{ n(L) - o(L) }, { n(L) - o(L) }>; m(L)] =
            decode_matrices::<{ n(L) - o(L) }, { n(L) - o(L) }, { m(L) }, { L }>(
                &p[0..p1_bytes(L)],
                m(L),
                true,
            );
        let p2: [Mat<{ n(L) - o(L) }, { o(L) }>; m(L)] =
            decode_matrices::<{ n(L) - o(L) }, { o(L) }, { m(L) }, { L }>(
                &p[p1_bytes(L)..(p1_bytes(L) + p2_bytes(L))],
                m(L),
                false,
            );

        let p3 = (0..m(L))
            .into_par_iter()
            .map(|i| upper(-om.tp().mul_right(p1[i].mul_right(om)) + (-om.tp().mul_right(p2[i]))))
            .collect::<Vec<Mat<{ o(L) }, { o(L) }>>>();

        let p3_enc = encode_matrices::<{ o(L) }, { o(L) }, { m(L) }, { L }>(&p3, true);
        let mut out = [0; pk_seed_bytes(L) + p3_bytes(L)];
        out[..pk_seed_bytes(L)].copy_from_slice(&seed_pk[..pk_seed_bytes(L)]);
        out[pk_seed_bytes(L)..].copy_from_slice(&p3_enc);
        SigningKey(seed_sk, VerifyKey(out))
    }
}

pub fn expand_sk<const L: Level>(
    csk: [u8; salt_bytes(L)],
) -> [u8; salt_bytes(L) + o_bytes(L) + p1_bytes(L) + l_bytes(L)]
where
    [(); salt_bytes(L)]:,
    [(); pk_seed_bytes(L) + p3_bytes(L)]:,
    [(); o_bytes(L) + pk_seed_bytes(L)]:,
    [(); n(L) - o(L)]:,
    [(); m(L)]:,
{
    let seed_sk = csk;

    let s: [u8; o_bytes(L) + pk_seed_bytes(L)] = shake_256(vec![&seed_sk]);

    let seed_pk = &s[0..pk_seed_bytes(L)];

    let o_bs = &s[pk_seed_bytes(L)..(pk_seed_bytes(L) + o_bytes(L))];
    let om: Mat<{ n(L) - o(L) }, { o(L) }> =
        decode_matrix::<{ n(L) - o(L) }, { o(L) }, { L }>(o_bs);

    let p = aes_128_ctr(seed_pk, p1_bytes(L) + p2_bytes(L));

    let p1: [Mat<{ n(L) - o(L) }, { n(L) - o(L) }>; m(L)] = decode_matrices::<
        { n(L) - o(L) },
        { n(L) - o(L) },
        { m(L) },
        { L },
    >(&p[0..p1_bytes(L)], m(L), true);

    let p1_bm = encode_matrices::<{ n(L) - o(L) }, { n(L) - o(L) }, { m(L) }, { L }>(&p1, true);

    let p2: [Mat<{ n(L) - o(L) }, { o(L) }>; m(L)] =
        decode_matrices::<{ n(L) - o(L) }, { o(L) }, { m(L) }, { L }>(
            &p[p1_bytes(L)..(p1_bytes(L) + p2_bytes(L))],
            m(L),
            false,
        );

    let l = (0..m(L))
        .into_par_iter()
        .map(|i| (p1[i] + p1[i].tp()).mul_right(om) + p2[i])
        .collect::<Vec<Mat<{ n(L) - o(L) }, { o(L) }>>>();

    let l_bs = encode_matrices::<{ n(L) - o(L) }, { o(L) }, { m(L) }, { L }>(&l, false);
    let mut out = [0; salt_bytes(L) + o_bytes(L) + p1_bytes(L) + l_bytes(L)];
    out[..salt_bytes(L)].copy_from_slice(&seed_sk);
    out[salt_bytes(L)..(salt_bytes(L) + o_bytes(L))].copy_from_slice(o_bs);
    out[(salt_bytes(L) + o_bytes(L))..(salt_bytes(L) + o_bytes(L) + p1_bytes(L))]
        .copy_from_slice(&p1_bm);
    out[(salt_bytes(L) + o_bytes(L) + p1_bytes(L))..].copy_from_slice(&l_bs);
    out
}

impl<const L: Level> Signer<Vec<u8>> for SigningKey<L>
where
    [(); salt_bytes(L)]:,
    [(); pk_seed_bytes(L) + p3_bytes(L)]:,
    [(); o_bytes(L) + pk_seed_bytes(L)]:,
    [(); n(L) - o(L)]:,
    [(); k(L) * n(L)]:,
    [(); m(L)]:,
    [(); m(L) / 2]:,
    [(); (k(L) * n(L) + 1) / 2]:,
    [(); k(L) * v_bytes(L) + ((k(L) * o(L) + 1) / 2)]:,
    [(); salt_bytes(L) + o_bytes(L) + p1_bytes(L) + l_bytes(L)]:,
    [(); digest_bytes(L)]:,
{
    fn sign(&self, msg: &[u8]) -> Vec<u8> {
        let esk = expand_sk(self.0);
        let seed_sk = &esk[..salt_bytes(L)];
        let om: Mat<{ n(L) - o(L) }, { o(L) }> = decode_matrix::<{ n(L) - o(L) }, { o(L) }, { L }>(
            &esk[salt_bytes(L)..(salt_bytes(L) + o_bytes(L))],
        );
        let p1: [Mat<{ n(L) - o(L) }, { n(L) - o(L) }>; m(L)] =
            decode_matrices::<{ n(L) - o(L) }, { n(L) - o(L) }, { m(L) }, { L }>(
                &esk[(salt_bytes(L) + o_bytes(L))..(salt_bytes(L) + o_bytes(L) + p1_bytes(L))],
                m(L),
                true,
            );
        let li: [Mat<{ n(L) - o(L) }, { o(L) }>; m(L)] =
            decode_matrices::<{ n(L) - o(L) }, { o(L) }, { m(L) }, { L }>(
                &esk[(salt_bytes(L) + o_bytes(L) + p1_bytes(L))..],
                m(L),
                false,
            );

        let m_digest: [u8; digest_bytes(L)] = shake_256(vec![msg]);

        let salt: [u8; salt_bytes(L)] = shake_256(vec![&m_digest, seed_sk]);

        let t_bs: [u8; m(L) / 2] = shake_256(vec![&m_digest, &salt]);

        let mut t = decode_vec::<{ m(L) }>(&t_bs);
        let v: [u8; k(L) * v_bytes(L) + ((k(L) * o(L) + 1) / 2)] =
            shake_256(vec![&m_digest, &salt, seed_sk, &[128]]);
        let mut vv: Vec<Mat<{ n(L) - o(L) }, 1>> = vec![];
        for i in 0..k(L) {
            vv.push(decode_vec::<{ n(L) - o(L) }>(
                &v[(i * v_bytes(L))..((i + 1) * v_bytes(L))],
            ));
        }
        let r = decode_vec::<{ k(L) * o(L) }>(&v[(k(L) * v_bytes(L))..]);

        let mut A = Mat([[0; { m(L) }]; { k(L) * o(L) }]);

        let mut l = 0;

        let mut el = Mat([[0; m(L)]; m(L)]);
        for i in 0..m(L) {
            el.write(i, i, 1);
        }

        let mut Mi = [Mat([[0; { o(L) }]; { m(L) }]); k(L)];

        for i in 0..k(L) {
            Mi[i] = Mat([[0; { o(L) }]; { m(L) }]);
            for j in 0..m(L) {
                Mi[i].0[j] = (vv[i].tp().mul_right(li[i])).0[0];
            }
        }

        for i in 0..k(L) {
            for j in (i..(k(L))).rev() {
                let mut u: Mat<{ m(L) }, 1> = Mat([[0; 1]; m(L)]);
                if j == i {
                    for a in 0..m(L) {
                        u.write(a, 0, vv[i].tp().mul_right(p1[a]).mul_right(vv[i]).0[0][0]);
                    }
                } else {
                    for a in 0..m(L) {
                        u.write(
                            a,
                            0,
                            (vv[i].tp().mul_right(p1[a]).mul_right(vv[j])
                                + vv[j].tp().mul_right(p1[a]).mul_right(vv[i]))
                            .0[0][0],
                        );
                    }
                }
                t = t + (-el.mul_right(u));
                for z in (i * o(L))..((i + 1) * o(L)) {
                    for y in 0..(m(L)) {
                        A.write(y, z, A.access(y, z) + el.mul_right(Mi[j]).0[y][z]);
                    }
                }

                if i != j {
                    for z in (j * o(L))..((j + 1) * o(L)) {
                        for y in 0..(m(L)) {
                            A.write(y, z, A.access(y, z) + el.mul_right(Mi[i]).0[y][z]);
                        }
                    }
                }
                l += 1;
                el = el.mul_right(Mat(polymat()));
            }
        }

        let x = sample_solution(A.tp(), t, r).tp();

        let mut s = Mat([[0; k(L) * n(L)]; 1]);
        for i in 0..(k(L)) {
            let x_inner = &x.0[0][(i * o(L))..((i + 1) * o(L))];
            let mut xi = [0; o(L)];
            for i in 0..o(L) {
                xi[i] = x_inner[i];
            }
            s.0[0][(i * n(L))..((i + 1) * n(L) - o(L))]
                .copy_from_slice(&(vv[i] + (om.mul_right(Mat([xi; 1]).tp()))).0[0]);
            s.0[0][((i + 1) * n(L) - o(L))..((i + 1) * n(L))]
                .copy_from_slice(&x.0[0][(i * o(L))..((i + 1) * o(L))]);
        }

        let mut sig = encode_vec::<{ k(L) * n(L) }>(s.tp()).to_vec();

        for i in salt {
            sig.push(i);
        }

        return sig;
    }
}

impl<const L: Level> Verifier<Vec<u8>> for VerifyKey<L>
where
    [(); pk_seed_bytes(L) + p3_bytes(L)]:,
{
    fn verify(&self, sig: &Vec<u8>, msg: &[u8]) -> bool {
        todo!();
    }
}

#[cfg(test)]
mod tests {
    use super::SigningKey;
    use crate::params::Level;
    use crate::sig::KeyGen;

    #[test]
    fn test_gen() {
        let s: SigningKey<{ Level::MAYO1 }> = SigningKey::gen();
        println!("{:?}", s);
    }
}
